/**
 * @file SeedService_service.cc
 * @brief Assists in the distribution of guaranteed unique seeds to all engines within a job.
 * @author Rob Kutschke (kutschke@fnal.gov)
 * @see SeedService.hh SeedMaster.hh
 */

// SeedService header
#include "artextensions/SeedService/SeedService.hh"

// Art include files
#include "art/Utilities/Exception.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Services/Registry/ActivityRegistry.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Persistency/Provenance/ModuleDescription.h"
#include "art/Persistency/Provenance/EventID.h"

// Supporting library include files
#include "messagefacility/MessageLogger/MessageLogger.h"

// CLHEP libraries
#include "CLHEP/Random/RandomEngine.h" // CLHEP::HepRandomEngine

// C++ include files
#include <iostream>
#include <iomanip>

namespace artext {

  //----------------------------------------------------------------------------
  SeedService::SeedService
    (fhicl::ParameterSet const& paramSet, art::ActivityRegistry& iRegistry)
    : seeds(paramSet)
    , state()
    , verbosity(paramSet.get<int>("verbosity", 0))
    , bPrintEndOfJobSummary(paramSet.get<bool>("endOfJobSummary",false))
  {
    // Register callbacks.
    iRegistry.sPreModuleConstruction.watch  (this, &SeedService::preModuleConstruction  );
    iRegistry.sPostModuleConstruction.watch (this, &SeedService::postModuleConstruction );
    iRegistry.sPreModuleBeginRun.watch      (this, &SeedService::preModuleBeginRun      );
    iRegistry.sPostModuleBeginRun.watch     (this, &SeedService::postModuleBeginRun     );
    iRegistry.sPreProcessEvent.watch        (this, &SeedService::preProcessEvent        );
    iRegistry.sPreModule.watch              (this, &SeedService::preModule              );
    iRegistry.sPostModule.watch             (this, &SeedService::postModule             );
    iRegistry.sPostProcessEvent.watch       (this, &SeedService::postProcessEvent       );
    iRegistry.sPostEndJob.watch             (this, &SeedService::postEndJob             );
    
  } // SeedService::SeedService()
  
  
  
  //----------------------------------------------------------------------------
  SeedService::EngineId SeedService::qualify_engine_label
    (std::string moduleLabel, std::string instanceName) const
    { return { moduleLabel, instanceName }; }
  
  SeedService::EngineId SeedService::qualify_engine_label
    (std::string instanceName /* = "" */) const
    { return qualify_engine_label( state.moduleLabel(), instanceName); }
  
  //----------------------------------------------------------------------------
  SeedService::seed_t SeedService::getSeed() {
    return getSeed(qualify_engine_label());
  } // SeedService::getSeed()
  
  
  //----------------------------------------------------------------------------
  SeedService::seed_t SeedService::getSeed(std::string instanceName) {
    return getSeed(qualify_engine_label(instanceName));
  } // SeedService::getSeed(string)
  
  
  //----------------------------------------------------------------------------
  void SeedService::print() const { seeds.print(); }
  
  
  //----------------------------------------------------------------------------
  SeedService::seed_t SeedService::getSeed(EngineId const& id) {
    
    // Are we being called from the right place?
    ensureValidState();
    
    // Ask the seed master about seeds...
    return seeds.getSeed(id);
  } // SeedService::getSeed(EngineID)
  
  
  std::pair<SeedService::seed_t, bool> SeedService::getSeed(
    EngineId const& id,
    fhicl::ParameterSet const& pset, std::initializer_list<std::string> pname
  ) {
    seed_t seed = InvalidSeed;
    bool bFrozen = readSeedParameter(seed, pset, pname);
    // if we got a seed, but it is invalid, we pretend we got nothing
    if (bFrozen && (seed == InvalidSeed)) bFrozen = false;
    return { (bFrozen? seed: getSeed(id)), bFrozen };
  } // SeedService::getSeed(ParameterSet)
  
  
  //----------------------------------------------------------------------------
  SeedService::seed_t SeedService::createEngine(
    art::EngineCreator& module, std::string type,
    std::string instance /* = "" */
  ) {
    EngineId id = qualify_engine_label(instance);
    const seed_t seed = prepareEngine(id, RandomNumberGeneratorSeeder);
    module.createEngine(seed, type, instance);
    mf::LogInfo("SeedService")
      << "Seeding " << type << " engine \"" << id.artName()
      << "\" with seed " << seed << ".";
    return seed;
  } // SeedService::createEngine(strings)
  
  
  SeedService::seed_t SeedService::createEngine(art::EngineCreator& module) {
    EngineId id = qualify_engine_label();
    const seed_t seed = prepareEngine(id, RandomNumberGeneratorSeeder);
    module.createEngine(seed);
    mf::LogInfo("SeedService")
      << "Seeding default-type engine \"" << id.artName()
      << "\" with seed " << seed << ".";
    return seed;
  } // SeedService::createEngine()
  
  
  SeedService::seed_t SeedService::createEngine(
    art::EngineCreator& module, std::string type,
    std::string instance,
    fhicl::ParameterSet const& pset, std::initializer_list<std::string> pname
  ) {
    EngineId id = qualify_engine_label(instance);
    registerEngineAndSeeder(id, RandomNumberGeneratorSeeder);
    std::pair<seed_t, bool> seedInfo = getSeed(id, pset, pname);
    module.createEngine(seedInfo.first, type, instance);
    mf::LogInfo("SeedService")
      << "Seeding " << type << " engine \"" << id.artName()
      << "\" with seed " << seedInfo.first << ".";
    if (seedInfo.second) freezeSeed(id);
    return seedInfo.first;
  } // SeedService::createEngine(ParameterSet)
  
  
  SeedService::seed_t SeedService::createEngine(
    art::EngineCreator& module,
    fhicl::ParameterSet const& pset, std::initializer_list<std::string> pname
  ) {
    EngineId id = qualify_engine_label();
    registerEngineAndSeeder(id, RandomNumberGeneratorSeeder);
    std::pair<seed_t, bool> seedInfo = getSeed(id, pset, pname);
    module.createEngine(seedInfo.first);
    mf::LogInfo("SeedService")
      << "Seeding default-type engine \"" << id.artName()
      << "\" with seed " << seedInfo.first << ".";
    if (seedInfo.second) freezeSeed(id);
    return seedInfo.first;
  } // SeedService::createEngine(ParameterSet)
  
  
  SeedService::seed_t SeedService::registerEngine
    (SeedMaster_t::Seeder_t seeder, std::string instance /* = "" */)
  {
    EngineId id = qualify_engine_label(instance);
    const seed_t seed = prepareEngine(id, seeder);
    if (seeder) {
      seeder(id, seed);
      mf::LogInfo("SeedService")
        << "Seeding registered engine \"" << id.artName()
        << "\" with seed " << seed << ".";
    }
    return seed;
  } // SeedService::registerEngine()
  
  
  SeedService::seed_t SeedService::registerEngine(
    SeedMaster_t::Seeder_t seeder, std::string instance,
    fhicl::ParameterSet const& pset, std::initializer_list<std::string> pname
  ) {
    EngineId id = qualify_engine_label(instance);
    registerEngineAndSeeder(id, seeder);
    std::pair<seed_t, bool> seedInfo = getSeed(id, pset, pname);
    if (seeder) {
      seeder(id, seedInfo.first);
      mf::LogInfo("SeedService")
        << "Seeding registered engine \"" << id.artName()
        << "\" with seed " << seedInfo.first << ".";
    }
    if (seedInfo.second) freezeSeed(id);
    return seedInfo.first;
  } // SeedService::registerEngine()
  
  
  //----------------------------------------------------------------------------
  void SeedService::ensureValidState() const {
    // getSeed may only be called from a c'tor or from a beginRun method.
    // In all other cases, throw.
    if ( (state.state() != SeedServiceHelper::ArtState::inModuleConstructor)
      && (state.state() != SeedServiceHelper::ArtState::inBeginRun)
      )
    {
      throw art::Exception(art::errors::LogicError)
        << "SeedService: not in a module constructor or beginRun method. May not call getSeed.";
    }
  } // SeedService::ensureValidState()
  
  
  //----------------------------------------------------------------------------
  SeedService::seed_t SeedService::reseedInstance(EngineId const& id) {
    // get all the information on the current process, event and module from
    // ArtState:
    SeedMaster_t::EventData_t data(state.getEventSeedInputData());
    seed_t seed = seeds.reseedEvent(id, data);
    if (seed == InvalidSeed) {
      mf::LogDebug("SeedService")
        << "No random seed specific to this event for engine '" << id << "'";
    }
    else {
      mf::LogInfo("SeedService") << "Random seed for this event, engine '"
        << id << "': " << seed;
    }
    return seed;
  } // SeedService::reseedInstance()
  
  
  void SeedService::reseedModule(std::string currentModule) {
    for (EngineId const& ID: seeds.engineIDsRange()) {
      if (ID.moduleLabel != currentModule) continue; // not our module? neeext!!
      reseedInstance(ID);
    } // for
  } // SeedService::reseedModule(string)
  
  void SeedService::reseedModule() { reseedModule(state.moduleLabel()); }
  
  
  //----------------------------------------------------------------------------
  void SeedService::RandomNumberGeneratorSeeder
    (EngineId const& id, seed_t seed)
  {
    // no check is performed to verify that the current module is the one
    // specified in id.moduleLabel -- but that is required!
    art::ServiceHandle<art::RandomNumberGenerator> rng;
    CLHEP::HepRandomEngine& engine = rng->getEngine(id.instanceName);
    engine.setSeed(seed, 0); // the 0 is dummy... or so one hopes
  } // SeedService::RandomNumberGeneratorSeeder()
  
  
  //----------------------------------------------------------------------------
  void SeedService::registerEngineAndSeeder
    (EngineId const& id, SeedMaster_t::Seeder_t seeder)
  {
    if (hasEngine(id)) {
      throw art::Exception(art::errors::LogicError)
        << "SeedService: an engine with ID '" << id
        << "' has already been created!";
    }
    seeds.registerSeeder(id, seeder);
  } // SeedService::registerEngine()
  
  
  //----------------------------------------------------------------------------
  void SeedService::freezeSeed(EngineId const& id)
    { seeds.freezeSeed(id); }
  
  
  //----------------------------------------------------------------------------
  SeedService::seed_t SeedService::prepareEngine
    (EngineId const& id, SeedMaster_t::Seeder_t seeder)
  {
    registerEngineAndSeeder(id, seeder);
    return getSeed(id);
  } // SeedService::prepareEngine()
  
  
  //----------------------------------------------------------------------------
  bool SeedService::readSeedParameter(
    seed_t& seed, fhicl::ParameterSet const& pset,
    std::initializer_list<std::string> pnames
  ) {
    for (std::string const& key: pnames)
      if (pset.get_if_present(key, seed)) return true;
    seed = InvalidSeed;
    return false;
  } // SeedService::readSeedParameter()
  
  
  //----------------------------------------------------------------------------
  // Callbacks called by art.  Used to maintain information about state.
  void SeedService::preModuleConstruction(art::ModuleDescription const& md)  {
    state.set_state(SeedServiceHelper::ArtState::inModuleConstructor);
    state.set_module(md);
  } // SeedService::preModuleConstruction()
  
  void SeedService::postModuleConstruction(art::ModuleDescription const&) {
    state.reset_state();
  } // SeedService::postModuleConstruction()
  
  void SeedService::preModuleBeginRun(art::ModuleDescription const& md) {
    state.set_state(SeedServiceHelper::ArtState::inBeginRun);
    state.set_module(md);
  } // SeedService::preModuleBeginRun()
  
  void SeedService::postModuleBeginRun(art::ModuleDescription const&) {
    state.reset_state();
  } // SeedService::postModuleBeginRun()
  
  void SeedService::preProcessEvent(art::Event const& evt) {
    state.set_state(SeedServiceHelper::ArtState::inEvent);
    state.set_event(evt);
    seeds.onNewEvent(); // inform the seed master that a new event has come
  } // SeedService::preProcessEvent()
  
  void SeedService::preModule(art::ModuleDescription const& md) {
    state.set_state(SeedServiceHelper::ArtState::inModuleEvent);
    state.set_module(md);
    
    // Reseed all the engine of this module... maybe
    // (that is, if the current policy alows it).
    reseedModule(md.moduleLabel());
  } // SeedService::preModule()
  
  void SeedService::postModule(art::ModuleDescription const&) {
    state.reset_module();
    state.reset_state();
  } // SeedService::postModule()
  
  void SeedService::postProcessEvent(art::Event const&) {
    state.reset_event();
    state.reset_state();
  } // SeedService::postProcessEvent()
  
  
  void SeedService::postEndJob() {
    if ((verbosity > 0) || bPrintEndOfJobSummary)
      print(); // framework logger decides whether and where it shows up
  } // SeedService::postEndJob()
  
  //----------------------------------------------------------------------------

} // end namespace artext

DEFINE_ART_SERVICE(artext::SeedService)
